/*
Programmed by John Palmisano, author of SocietyofRobots.com, 2010.

Commercial use of this code with products other than from societyofrobots.com is absolutely prohibited.

License grants permission to only use this code and derivitives with the Axon Mote microcontroller.
WebbotLib is granted special unlimited persmission to use this code.

Non-commercial use is allowed, however this header must always be included.

For more info, see iom128rfa1.h in the WinAVR includes
*/


//my defines
//#define TEST_OUTPUT	//use for debugging wireless code, only
#define WIRELESS_SLEEP	1
#define WIRELESS_WAKE	0

//wireless frame stored buffer
#define MAX_FRAME_SIZE (127)
char wireless_recieved_frame[MAX_FRAME_SIZE];

//transmission quality variables, only updated during recieved frames
uint8_t wireless_random_number;
uint8_t wireless_RSSI_val=0;
uint8_t wireless_ED_measurement=0;
uint8_t wireless_LQI=0;
static volatile boolean wireless_crc_valid=false;
static volatile boolean battery_low=false;

//states
static volatile boolean tx_in_progress=false;
static volatile boolean pll_locked=false;

//misc variables
static uint8_t low_voltage_threshold=10;//at 10, with HR=1, LOW_BAT interrupt will only trigger if below 3.3V
uint8_t frame_size=0;//keeps track of frame size to prevent overwriting of wireless data buffer before xmit


//enable the interrupts and other settings that are important to Axon Mote
void wireless_Enable_Interrupts(void)
	{
	//disable antenna diversity
	ANT_DIV_struct.ant_div_en=0;

	//enable interrupts
	IRQ_MASK_struct.pll_lock_en=1;
	IRQ_MASK_struct.pll_unlock_en=1;
	IRQ_MASK_struct.rx_start_en=1;
	IRQ_MASK_struct.rx_end_en=1;
	IRQ_MASK_struct.tx_end_en=1;
	IRQ_MASK_struct.cca_ed_done_en=1;
	//IRQ_MASK_struct.awake_en=1;

	//battery monitor interrupt settings - note that low battery sometimes triggers when writing to uSD
	BATMON_struct.bat_low_en=0;
	BATMON_struct.batmon_hr=BATMON_HR_EN;
	BATMON_struct.batmon_vth=low_voltage_threshold;

	//enable automatic CRC calculation
	TRX_CTRL_1_struct.tx_auto_crc_on=1;

	//set TX power
	PHY_TX_PWR_struct.tx_pwr=0;//from 0 (max) to 15 (min)
	PHY_TX_PWR_struct.pa_lt=PA_LT_8US;
	PHY_TX_PWR_struct.pa_buf_lt=PA_BUF_LT_6US;

	//set data rate, options are 250, 500, 1000, and 2000 - see page 86 of datasheet before changing
	TRX_CTRL_2_struct.oqpsk_data_rate=RATE_250KB;
	}

//calling this functiton will create a delay until a transistion is detected
//never do anything when states are being transitioned
void wireless_transition_wait(void)
	{
	//wait until transistion is finished
	while(TRX_STATUS_struct.trx_status == STATE_TRANSITION_IN_PROGRESS)
		{
		#ifdef TEST_OUTPUT
			rprintf("transition detected");
		#endif
		}
	}

/*
Reset Transceiver

A radio transceiver reset is initiated by setting bit
TRXRST in register TRXPR. This bit is self-resetting.

After RESET all interrupts are disabled. During radio transceiver initialization it is 
recommended to enable AWAKE to be notified once the TRX_OFF state is entered. 
Note that the TRX24_AWAKE interrupt can usually not be seen when the transceiver 
enters TRX_OFF state after RESET, because register  IRQ_MASK is reset to mask all 
interrupts. In this case, state TRX_OFF is normally entered before the microcontroller 
could modify the register.

Required interrupts must be enabled by writing to register IRQ_MASK and the global 
interrupt enable flag must be set.
*/
void wireless_reset(void)
	{
	wireless_transition_wait();//if transitioning, wait

	//"During radio transceiver reset the TRXPR register is not cleared
	//and therefore the application software has to set the SLPTR bit to 0."
	TRXPR &= ~(1 << SLPTR);

	wireless_transition_wait();//if transitioning, wait
	
	TRXPR |= (1 << TRXRST);//reset

	//wait for reset to complete
	while(TRX_STATUS_struct.trx_status != TRX_OFF);

	#ifdef TEST_OUTPUT
	if(TRX_STATUS_struct.trx_status == TRX_OFF)
		rprintf("\nreset detected");
	#endif
	
	//when reset, the interrupt bits get erased
	wireless_Enable_Interrupts();//set the interrupt bits on that we want
	sei();//turn on interrupts (if not already on)
	}


//go to TRX_OFF mode
void wireless_TRX_OFF(void)
	{
	TRX_STATE_struct.trx_cmd = CMD_TRX_OFF;
	
	#ifdef TEST_OUTPUT
	while(TRX_STATUS_struct.trx_status != TRX_OFF);
	if(TRX_STATUS_struct.trx_status == TRX_OFF)
		rprintf("\nTRX now OFF");
	#endif
	}

/*
Sleep mode
-hasn't been verified, yet.

The radio transceiver signal SLPTR can be controlled by
the bit SLPTR in register TRXPR and is used to set the
radio transceiver into SLEEP mode (assuming TRX_STATE is TRX_OFF).
set 1 to sleep and 0 to wake up.

 
If the radio transceiver was in SLEEP state, the SLPTR bit in the TRXPR register must 
be cleared prior to clearing the TRXRST bit in order to enter the TRX_OFF state. 
Otherwise the radio transceiver enters the SLEEP state immediately.

from WebbotLib:
For SLEEP then the current state MUST already be TRX_OFF.
Then do:
TRXPR_struct.slptr=1; // enter sleep mode


To wake it up then:
TRXPR_struct.slptr=0;
should put it back into TRX_OFF mode
*/
void wireless_sleep(boolean state)
	{
	wireless_transition_wait();//if transitioning, wait

	//if told to go to sleep
	if(state==WIRELESS_SLEEP)
		{
		#ifdef TEST_OUTPUT
			rprintf("\nentering sleep mode from ");

			//if coming from RX_ON state
			if(TRX_STATUS_struct.trx_status == RX_ON)
				{
				rprintf("RX_ON");
				}

			//if coming from PLL_ON state
			else if(TRX_STATUS_struct.trx_status == PLL_ON)
				{
				rprintf("PLL_ON");
				}
		#endif

		//go into TRX_OFF mode first
		if(TRX_STATUS_struct.trx_status != TRX_OFF)
			{
			wireless_TRX_OFF();
			wireless_transition_wait();
			}

		//now that its at TRX_OFF, put it to sleep for good hehehe
		//TRXPR |= (1 << SLPTR);//sbi go to sleep
		//SMCR_struct.se=1;
		TRXPR_struct.slptr=1; // enter sleep mode
		
		#ifdef TEST_OUTPUT
			rprintf("\nprobably in sleep state");//I think the register doesn't work in SLEEP mode, so no way to check
		#endif
		}

	//if told to wake up
	else if(state==WIRELESS_WAKE)
		{
		//TRXPR &= ~(1 << SLPTR);//cbi, wake up
		//SMCR_struct.se=0;
		TRXPR_struct.slptr=0; // wake up
		
		#ifdef TEST_OUTPUT
			while(TRX_STATUS_struct.trx_status == SLEEP)
				rprintf("\nwake state detected, now in OFF mode");
		#endif
		}
	}

//Set Channel, ranges from 11 to 26
void wireless_SetChannel(uint8_t pchannel)
	{
	//stupid check
	if(pchannel < 11)
		pchannel=11;
	else if(pchannel > 26)
		pchannel=26;

	//set channel
	PHY_CC_CCA_struct.channel=pchannel;

	delay_us(25);//max settling time between channel switch
	
	#ifdef TEST_OUTPUT
		rprintf("\nnew channel selected");
	#endif
	}

/*
The CCA modes are configurable via register PHY_CC_CCA. 
Usimg the Basic Operating Mode, a CCA request can be initiated manually by setting 
CCA_REQUEST = 1 of register PHY_CC_CCA, if the radio transceiver is in any RX 
state. The current channel status (CCA_STATUS) and  the CCA completion status 
(CCA_DONE) are accessible in register TRX_STATUS.

The CCA evaluation is done over eight symbol periods and the result is accessible 
tTR28 = 140 s (128 s measurement duration and processing delay) after the request. 
*/
uint8_t wireless_ReadChannel(void)
	{
	uint8_t pchannel = 0;

	//initiate CCA request
	PHY_CC_CCA |= (1 << CCA_REQUEST);//sbi

	//wait until done
	delay_us(140);

	//read the channel
	pchannel=PHY_CC_CCA_struct.channel;
	
	#ifdef TEST_OUTPUT
	rprintf("\nchannel #: %d",pchannel);
	#endif

	return pchannel;
	}

/*
Entering the PLL_ON state from TRX_OFF state first  enables the analog voltage 
regulator (AVREG). After the voltage regulator has  been settled the PLL frequency 
synthesizer is enabled. When the PLL has been settled at the receive frequency to a 
channel defined by bits CHANNEL of register PHY_CC_CCA a successful PLL lock is 
indicated by issuing a TRX24_PLL_LOCK interrupt.

If an RX_ON command is issued in PLL_ON state, the receiver is immediately enabled. 
If the PLL has not been settled before the state change nevertheless takes place. Even 
if the register bits TRX_STATUS of register TRX_STATUS indicates RX_ON, actual 
frame reception can only start once the PLL has locked.
*/
void wireless_PLL_ON(void)
	{
	wireless_transition_wait();//if transitioning, wait
	//pll_locked=false;//it's not yet locked

	#ifdef TEST_OUTPUT
	if(TRX_STATUS_struct.trx_status != PLL_ON)
		rprintf("\nPLL is not on");

	//if coming from TRX_OFF state
	if(TRX_STATUS_struct.trx_status == TRX_OFF)
		rprintf("\ncoming from TRX_OFF");

	//if coming from RX_ON state
	else if(TRX_STATUS_struct.trx_status == RX_ON)
		rprintf("\ncoming from RX_ON");
	#endif

	TRX_STATE_struct.trx_cmd = CMD_PLL_ON;

	//time to transistion is much shorter than PLL Lock time, so must delay
	delay_us(110);
	//while(pll_locked!=true);

	/*while(TRX_STATUS_struct.trx_status != PLL_ON);

	#ifdef TEST_OUTPUT
	if(TRX_STATUS_struct.trx_status == PLL_ON)
		rprintf("\nPLL now ON");
	#endif*/
	
	//won't work if there isn't a lock
	while(IRQ_STATUS_struct.pll_lock != PLL_LOCK);
	}

//go into RX_ON (Rx Listen State)
/*
During RX_ON state the receiver listens for incoming frames. After detecting a valid 
synchronization header (SHR), the receiver automatically enters the BUSY_RX state. 
The reception of a valid PHY header (PHR) generates an TRX24_RX_START interrupt 
and receives and demodulates the PSDU data.

During PSDU reception the frame data are stored continuously in the Frame Buffer until 
the last byte was received. The completion of the frame reception is indicated by an 
TRX24_RX_END interrupt and the radio transceiver reenters the state RX_ON. At the 
same time the bits RX_CRC_VALID of register PHY_RSSI are updated with the result 
of the FCS check (see "Frame Check Sequence (FCS)" on page 66).
*/
void wireless_RX_ON(void)
	{
	wireless_transition_wait();//if transitioning, wait

	//if coming from TRX_OFF state
	if(TRX_STATUS_struct.trx_status == TRX_OFF)
		{
		pll_locked=false;

		TRX_STATE_struct.trx_cmd = CMD_RX_ON;
		
		//must wait until PLL is locked
		while(!pll_locked);

		#ifdef TEST_OUTPUT
		rprintf("\ncoming from TRX_OFF");
		#endif
		}

	//if coming from PLL_ON state
	else if(TRX_STATUS_struct.trx_status == PLL_ON)
		{
		TRX_STATE_struct.trx_cmd = CMD_RX_ON;

		#ifdef TEST_OUTPUT
		rprintf("\ncoming from PLL_ON");
		#endif
		}
	
	//wait until its on
	while(TRX_STATUS_struct.trx_status != RX_ON);

	#ifdef TEST_OUTPUT
	if(TRX_STATUS_struct.trx_status == RX_ON)
		rprintf("\nRX now ON");
	#endif
	}

//go into TX_ON (Tx Transmit State)
//running this function causes data in the buffer to transmit wirelessly
void wireless_BUSY_TX(void)
	{
	wireless_transition_wait();//if transitioning, wait

	tx_in_progress=true;

	//if coming from PLL_ON state
	if(TRX_STATUS_struct.trx_status == PLL_ON)
		{
		TRX_STATE_struct.trx_cmd = CMD_TX_START;
		
		//wait until its ready to transmit
		while(tx_in_progress);

		//reset frame counter
		frame_size=0;

		#ifdef TEST_OUTPUT
			rprintf("\ntransmission sent");
		#endif
		}
	//A transmission can only be initiated in state PLL_ON
	else
		rprintf("\ntransmission failed, must be in PLL_ON state\n");

	//wait until transmission is done
	wireless_transition_wait();//if transitioning, wait

	//return to PLL_ON
	TRX_STATE_struct.trx_cmd = CMD_PLL_ON;
	}

//call this to transmit data
//it transmits stuff in the buffer right away
//beware that it uses the same buffer as the Rx
void wireless_BUFFER_byte(uint8_t data)
	{
	frame_size++;

	//number of frames
	TRXFBST=frame_size+2;//first byte is frame length, the actual frame, then another frame for LQI (hence +2)

	//actual data
	_SFR_MEM8(0x180+frame_size)=data;
	}

//prints a null-terminated string stored in RAM
//string length must be less <=125 bytes
void wireless_BUFFER_string(char str[])
	{
	// send a string stored in RAM
	if(str)
		{
		// print the string until a null-terminator
		while (*str)
			{
			frame_size++;
			
			//within max buffer size
			if(frame_size<126)
				_SFR_MEM8(0x180+frame_size)=*str++;
			
			//too much stuff in buffer, stop storing and warn user
			if(frame_size==126)
				rprintf("\ntransmit error: data exceeded 125 bytes");
			}
		}
		
	//store number of frames
	TRXFBST=frame_size+2;//first byte is frame length, then actual frame, then another frame for LQI (hence +2)
	}

//get RSSI and RND_VALUE
void wireless_update_measurements(void)
	{
	/*
	In Basic Operating Mode (NOT EXTENDED MODE) the RSSI value is valid in any receive state, and is updated 
	every tTR25 = 2 s to register PHY_RSSI.

	CAN ONLY BE READ AFTER RX_START INTERRUPT!!!
	*/
	wireless_RSSI_val=PHY_RSSI_struct.rssi;

	//get REAL random number for use in wireless encryption, etc
	wireless_random_number=PHY_RSSI_struct.rnd_value;
	}


//simplified easy to remember functions
void wireless_TX_MODE(void)//prepare to transmit data
	{
	wireless_PLL_ON();
	}
void wireless_TX_TRANSMIT(void)//send data in buffer
	{
	wireless_BUSY_TX();
	}
void wireless_RX_MODE(void)//prepare to receive data
	{
	wireless_RX_ON();
	}
void wireless_RX_PROTECT(boolean protect)//prevent overwriting of stored data
	{
	if(protect==1)
		wireless_PLL_ON();//transmit mode blocks RX writing to buffer
	else
		wireless_RX_ON();//RX now allowed to write to buffer
	}
void wireless_OFF(void)//turn transceiver off to reduce power (more power than SLEEP, but works much faster)
	{
	wireless_TRX_OFF();
	}
void wireless_SLEEP(void)//put transceiver into sleep mode
	{
	wireless_sleep(1);
	}
void wireless_WAKE(void)//wake up from sleep mode
	{
	wireless_sleep(0);
	}
void wireless_BUFFER_CLEAN(void)//clean out buffer data
	{
	//data still exists, but the frame count is zero so it's as if it doesn't
	TRXFBST=0;//clear TRX buffer frame
	frame_size=0;//reset local buffer frame
	}

//transmission beginning
//this interrupt needed to signal start of transmission
ISR(TRX24_RX_START_vect)
	{
	//get RSSI and RND_VALUE
	wireless_update_measurements();

	//start ED measurement here and read later
	//initiate ED measurement by writing an arbitrary value to register PHY_ED_LEVEL
	PHY_ED_LEVEL_struct.ed_level=ED_RESET;

	#ifdef TEST_OUTPUT
		rprintf("\nRX_START");
	#endif
	}

//this interrupt triggers when RX data arrives, if in RX_ON mode
//stores data into a permenant buffer
ISR(TRX24_RX_END_vect)
	{
	//enable buffer protection bits here?
	//TRX_STATE_struct.trx_cmd = CMD_PLL_ON;

	//CRC check
	if(PHY_RSSI_struct.rx_crc_valid == CRC_VALID)
		{
		uint8_t frame_length=0;
		uint8_t i=0;

		//get recieved frame length size
		frame_length=TST_RX_LENGTH_struct.rx_length;
		
		#ifdef TEST_OUTPUT
			rprintf("\nframe size:%d\ndata: ",frame_length);
		#endif

		//rprintf("data: %d %d %d %d %d %d",_SFR_MEM8(0x180),_SFR_MEM8(0x181),_SFR_MEM8(0x182),_SFR_MEM8(0x183),_SFR_MEM8(0x184),_SFR_MEM8(0x185));

		//read in frame from frame buffer
		for(i=0x0;i<frame_length-2;i=i+0x1)//first frame for length, last frame for LQI (hence -2)
			{
			//store data
			wireless_recieved_frame[i]=_SFR_MEM8(0x180+i);//read from start of frame buffer
			//wireless_recieved_frame[i]=TRXFBEND;//read from end of frame buffer

			//show the data
			#ifdef TEST_OUTPUT
				rprintf("%c",wireless_recieved_frame[i]);
			#endif
			}
		//determine link quality assessment (low values=bad, means high packet loss ratio)
		wireless_LQI=_SFR_MEM8(0x180+frame_length);//NOT SURE IF THIS IS WORKING, MUST TEST!
		
		#ifdef TEST_OUTPUT
			rprintf("\nLQI=%d",wireless_LQI);
		#endif

		wireless_crc_valid=true;//good data
		}

	//disable buffer protection bits here?

	else// if(PHY_RSSI_struct.rx_crc_valid == CRC_INVALID)
		{
		wireless_crc_valid=false;//bad data

		#ifdef TEST_OUTPUT
			rprintf("\ntransmission CRC failed - bad data");
		#endif
		}
	}

//this interrupt needed to signal end of transmission
ISR(TRX24_TX_END_vect)
	{
	//transmission completed
	tx_in_progress = false;
	
	#ifdef TEST_OUTPUT
		rprintf("\nTX_END int");
	#endif
	}

//indicates a PLL lock
ISR(TRX24_PLL_LOCK_vect)
	{
	pll_locked = true;
	
	#ifdef TEST_OUTPUT
		rprintf("\nPLL_LOCK int");
	#endif
	}

//indicates a PLL lock
ISR(TRX24_PLL_UNLOCK_vect)
	{
	pll_locked = false;
	
	#ifdef TEST_OUTPUT
		rprintf("\nPLL_UNLOCK int");
	#endif
	}

//battery monitor interrupt
//be aware that the uSD sometimes triggers the low battery interrupt
ISR(BAT_LOW_vect)
	{
	BATMON_struct.bat_low_en=0;//turn off interrupt

	battery_low=true;

	#ifdef TEST_OUTPUT
		INTERRUPTABLE_SECTION{
		rprintf("\n\nLOW BATTERY\n\n");
		delay_ms(1000);
		}
	#endif
	
	BATMON_struct.bat_low_en=1;//turn on interrupt
	}

//an averaged RSSI over time - use for Extended Mode
ISR(TRX24_CCA_ED_DONE_vect)
	{
	wireless_ED_measurement=PHY_ED_LEVEL_struct.ed_level;

	#ifdef TEST_OUTPUT
		rprintf("\nED_INT:%d", wireless_ED_measurement);
	#endif
	}
